@@ -1,5 +1,5 @@
 /* cp-hash.c  -- file copying (hash search routines)
-   Copyright (C) 89, 90, 91, 1995-2003 Free Software Foundation.
+   Copyright (C) 89, 90, 91, 1995-2005 Free Software Foundation.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,7 +13,7 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software Foundation,
-   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
    Written by Torbjorn Granlund, Sweden (tege@sics.se).
    Rewritten to use lib/hash.c by Jim Meyering.  */
@@ -31,7 +31,7 @@
 #include "cp-hash.h"
 
 /* Use ST_DEV and ST_INO as the key, FILENAME as the value.
-   These are used e.g., in copy.c to associate the destination path with
+   These are used e.g., in copy.c to associate the destination name with
    the source device/inode pair so that if we encounter a matching dev/ino
    pair in the source tree we can arrange to create a hard link between
    the corresponding names in the destination tree.  */
@@ -39,8 +39,8 @@ struct Src_to_dest
 {
   ino_t st_ino;
   dev_t st_dev;
-  /* Destination path name (of non-directory or pre-existing directory)
-     corresponding to the dev/ino of a copied file, or the destination path
+  /* Destination file name (of non-directory or pre-existing directory)
+     corresponding to the dev/ino of a copied file, or the destination file
      name corresponding to a dev/ino pair for a newly-created directory. */
   char *name;
 };
@@ -52,8 +52,8 @@ static Hash_table *src_to_dest;
 /* Initial size of the above hash table.  */
 #define INITIAL_TABLE_SIZE 103
 
-static unsigned int
-src_to_dest_hash (void const *x, unsigned int table_size)
+static size_t
+src_to_dest_hash (void const *x, size_t table_size)
 {
   struct Src_to_dest const *p = x;
 
@@ -83,7 +83,7 @@ src_to_dest_free (void *x)
 
 /* Remove the entry matching INO/DEV from the table
    that maps source ino/dev to destination file name.  */
-void
+extern void
 forget_created (ino_t ino, dev_t dev)
 {
   struct Src_to_dest probe;
@@ -98,29 +98,43 @@ forget_created (ino_t ino, dev_t dev)
     src_to_dest_free (ent);
 }
 
-/* Add PATH to the list of files that we have created.
-   Return 1 if we can't stat PATH, otherwise 0.  */
+/* Add FILE to the list of files that we have created.
+   Return true if successful.  */
 
-int
-remember_created (const char *path)
+extern bool
+remember_created (char const *file)
 {
   struct stat sb;
 
-  if (stat (path, &sb) < 0)
+  if (stat (file, &sb) < 0)
     {
-      error (0, errno, "%s", quote (path));
-      return 1;
+      error (0, errno, "%s", quote (file));
+      return false;
     }
 
-  remember_copied (path, sb.st_ino, sb.st_dev);
-  return 0;
+  remember_copied (file, sb.st_ino, sb.st_dev);
+  return true;
 }
 
-/* Add path NAME, copied from inode number INO and device number DEV,
+/* If INO/DEV correspond to an already-copied source file, return the
+   name of the corresponding destination file.  Otherwise, return NULL.  */
+
+extern char *
+src_to_dest_lookup (ino_t ino, dev_t dev)
+{
+  struct Src_to_dest ent;
+  struct Src_to_dest const *e;
+  ent.st_ino = ino;
+  ent.st_dev = dev;
+  e = hash_lookup (src_to_dest, &ent);
+  return e ? e->name : NULL;
+}
+
+/* Add file NAME, copied from inode number INO and device number DEV,
    to the list of files we have copied.
    Return NULL if inserted, otherwise non-NULL. */
 
-char *
+extern char *
 remember_copied (const char *name, ino_t ino, dev_t dev)
 {
   struct Src_to_dest *ent;
@@ -152,7 +166,7 @@ remember_copied (const char *name, ino_t ino, dev_t dev)
 }
 
 /* Initialize the hash table.  */
-void
+extern void
 hash_init (void)
 {
   src_to_dest = hash_initialize (INITIAL_TABLE_SIZE, NULL,
@@ -166,7 +180,7 @@ hash_init (void)
 /* Reset the hash structure in the global variable `htab' to
    contain no entries.  */
 
-void
+extern void
 forget_all (void)
 {
   hash_free (src_to_dest);
