@@ -1,5 +1,5 @@
 /* id -- print real and effective UIDs and GIDs
-   Copyright (C) 1989-2003 Free Software Foundation, Inc.
+   Copyright (C) 1989-2005 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,7 +13,7 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software Foundation,
-   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
+   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */
 
 /* Written by Arnold Robbins.
    Major rewrite by David MacKenzie, djm@gnu.ai.mit.edu. */
@@ -28,21 +28,12 @@
 
 #include "system.h"
 #include "error.h"
-#include "closeout.h"
+#include "quote.h"
 
 /* The official name of this program (e.g., no `g' prefix).  */
 #define PROGRAM_NAME "id"
 
-#define AUTHORS N_ ("Arnold Robbins and David MacKenzie")
-
-#ifndef _POSIX_VERSION
-struct passwd *getpwuid ();
-struct group *getgrgid ();
-uid_t getuid ();
-gid_t getgid ();
-uid_t geteuid ();
-gid_t getegid ();
-#endif /* not _POSIX_VERSION */
+#define AUTHORS "Arnold Robbins", "David MacKenzie"
 
 int getugroups ();
 
@@ -54,15 +45,15 @@ static void print_full_info (const char *username);
 /* The name this program was run with. */
 char *program_name;
 
-/* If nonzero, output user/group name instead of ID number. -n */
-static int use_name = 0;
+/* If true, output user/group name instead of ID number. -n */
+static bool use_name = false;
 
 /* The real and effective IDs of the user to print. */
 static uid_t ruid, euid;
 static gid_t rgid, egid;
 
-/* The number of errors encountered so far. */
-static int problems = 0;
+/* True unless errors have been encountered.  */
+static bool ok = true;
 
 static struct option const longopts[] =
 {
@@ -79,7 +70,7 @@ static struct option const longopts[] =
 void
 usage (int status)
 {
-  if (status != 0)
+  if (status != EXIT_SUCCESS)
     fprintf (stderr, _("Try `%s --help' for more information.\n"),
 	     program_name);
   else
@@ -111,14 +102,14 @@ main (int argc, char **argv)
 {
   int optc;
 
-  /* If nonzero, output the list of all group IDs. -G */
-  int just_group_list = 0;
-  /* If nonzero, output only the group ID(s). -g */
-  int just_group = 0;
-  /* If nonzero, output real UID/GID instead of default effective UID/GID. -r */
-  int use_real = 0;
-  /* If nonzero, output only the user ID(s). -u */
-  int just_user = 0;
+  /* If true, output the list of all group IDs. -G */
+  bool just_group_list = false;
+  /* If true, output only the group ID(s). -g */
+  bool just_group = false;
+  /* If true, output real UID/GID instead of default effective UID/GID. -r */
+  bool use_real = false;
+  /* If true, output only the user ID(s). -u */
+  bool just_user = false;
 
   initialize_main (&argc, &argv);
   program_name = argv[0];
@@ -132,25 +123,23 @@ main (int argc, char **argv)
     {
       switch (optc)
 	{
-	case 0:
-	  break;
 	case 'a':
 	  /* Ignore -a, for compatibility with SVR4.  */
 	  break;
 	case 'g':
-	  just_group = 1;
+	  just_group = true;
 	  break;
 	case 'n':
-	  use_name = 1;
+	  use_name = true;
 	  break;
 	case 'r':
-	  use_real = 1;
+	  use_real = true;
 	  break;
 	case 'u':
-	  just_user = 1;
+	  just_user = true;
 	  break;
 	case 'G':
-	  just_group_list = 1;
+	  just_group_list = true;
 	  break;
 	case_GETOPT_HELP_CHAR;
 	case_GETOPT_VERSION_CHAR (PROGRAM_NAME, AUTHORS);
@@ -162,12 +151,15 @@ main (int argc, char **argv)
   if (just_user + just_group + just_group_list > 1)
     error (EXIT_FAILURE, 0, _("cannot print only user and only group"));
 
-  if (just_user + just_group + just_group_list == 0 && (use_real || use_name))
+  if (just_user + just_group + just_group_list == 0 && (use_real | use_name))
     error (EXIT_FAILURE, 0,
 	   _("cannot print only names or real IDs in default format"));
 
   if (argc - optind > 1)
-    usage (EXIT_FAILURE);
+    {
+      error (0, 0, _("extra operand %s"), quote (argv[optind + 1]));
+      usage (EXIT_FAILURE);
+    }
 
   if (argc - optind == 1)
     {
@@ -195,7 +187,7 @@ main (int argc, char **argv)
     print_full_info (argv[optind]);
   putchar ('\n');
 
-  exit (problems != 0);
+  exit (ok ? EXIT_SUCCESS : EXIT_FAILURE);
 }
 
 /* Print the name or value of user ID UID. */
@@ -210,13 +202,14 @@ print_user (uid_t uid)
       pwd = getpwuid (uid);
       if (pwd == NULL)
 	{
-	  error (0, 0, _("cannot find name for user ID %u"), uid);
-	  problems++;
+	  error (0, 0, _("cannot find name for user ID %lu"),
+		 (unsigned long int) uid);
+	  ok = false;
 	}
     }
 
   if (pwd == NULL)
-    printf ("%u", (unsigned) uid);
+    printf ("%lu", (unsigned long int) uid);
   else
     printf ("%s", pwd->pw_name);
 }
@@ -233,13 +226,14 @@ print_group (gid_t gid)
       grp = getgrgid (gid);
       if (grp == NULL)
 	{
-	  error (0, 0, _("cannot find name for group ID %u"), gid);
-	  problems++;
+	  error (0, 0, _("cannot find name for group ID %lu"),
+		 (unsigned long int) gid);
+	  ok = false;
 	}
     }
 
   if (grp == NULL)
-    printf ("%u", (unsigned) gid);
+    printf ("%lu", (unsigned long int) gid);
   else
     printf ("%s", grp->gr_name);
 }
@@ -248,39 +242,42 @@ print_group (gid_t gid)
 
 /* FIXME: document */
 
-static int
+static bool
 xgetgroups (const char *username, gid_t gid, int *n_groups,
 	    GETGROUPS_T **groups)
 {
   int max_n_groups;
   int ng;
-  GETGROUPS_T *g;
-  int fail = 0;
+  GETGROUPS_T *g = NULL;
 
-  if (username == 0)
+  if (!username)
     max_n_groups = getgroups (0, NULL);
   else
     max_n_groups = getugroups (0, NULL, username, gid);
 
-  /* Add 1 just in case max_n_groups is zero.  */
-  g = xmalloc (max_n_groups * sizeof (GETGROUPS_T) + 1);
-  if (username == 0)
-    ng = getgroups (max_n_groups, g);
+  if (max_n_groups < 0)
+    ng = -1;
   else
-    ng = getugroups (max_n_groups, g, username, gid);
+    {
+      g = xnmalloc (max_n_groups, sizeof *g);
+      if (!username)
+	ng = getgroups (max_n_groups, g);
+      else
+	ng = getugroups (max_n_groups, g, username, gid);
+    }
 
   if (ng < 0)
     {
       error (0, errno, _("cannot get supplemental group list"));
-      ++fail;
-      free (groups);
+      free (g);
+      return false;
     }
-  if (!fail)
+  else
     {
       *n_groups = ng;
       *groups = g;
+      return true;
     }
-  return fail;
 }
 
 #endif /* HAVE_GETGROUPS */
@@ -294,7 +291,7 @@ print_group_list (const char *username)
 
   pwd = getpwuid (ruid);
   if (pwd == NULL)
-    problems++;
+    ok = false;
 
   print_group (rgid);
   if (egid != rgid)
@@ -307,12 +304,12 @@ print_group_list (const char *username)
   {
     int n_groups;
     GETGROUPS_T *groups;
-    register int i;
+    int i;
 
-    if (xgetgroups (username, (pwd ? pwd->pw_gid : (gid_t) -1),
-		    &n_groups, &groups))
+    if (! xgetgroups (username, (pwd ? pwd->pw_gid : (gid_t) -1),
+		      &n_groups, &groups))
       {
-	++problems;
+	ok = false;
 	return;
       }
 
@@ -335,37 +332,29 @@ print_full_info (const char *username)
   struct passwd *pwd;
   struct group *grp;
 
-  printf ("uid=%u", (unsigned) ruid);
+  printf ("uid=%lu", (unsigned long int) ruid);
   pwd = getpwuid (ruid);
-  if (pwd == NULL)
-    problems++;
-  else
+  if (pwd)
     printf ("(%s)", pwd->pw_name);
 
-  printf (" gid=%u", (unsigned) rgid);
+  printf (" gid=%lu", (unsigned long int) rgid);
   grp = getgrgid (rgid);
-  if (grp == NULL)
-    problems++;
-  else
+  if (grp)
     printf ("(%s)", grp->gr_name);
 
   if (euid != ruid)
     {
-      printf (" euid=%u", (unsigned) euid);
+      printf (" euid=%lu", (unsigned long int) euid);
       pwd = getpwuid (euid);
-      if (pwd == NULL)
-	problems++;
-      else
+      if (pwd)
 	printf ("(%s)", pwd->pw_name);
     }
 
   if (egid != rgid)
     {
-      printf (" egid=%u", (unsigned) egid);
+      printf (" egid=%lu", (unsigned long int) egid);
       grp = getgrgid (egid);
-      if (grp == NULL)
-	problems++;
-      else
+      if (grp)
 	printf ("(%s)", grp->gr_name);
     }
 
@@ -373,12 +362,12 @@ print_full_info (const char *username)
   {
     int n_groups;
     GETGROUPS_T *groups;
-    register int i;
+    int i;
 
-    if (xgetgroups (username, (pwd ? pwd->pw_gid : (gid_t) -1),
-		    &n_groups, &groups))
+    if (! xgetgroups (username, (pwd ? pwd->pw_gid : (gid_t) -1),
+		      &n_groups, &groups))
       {
-	++problems;
+	ok = false;
 	return;
       }
 
@@ -388,11 +377,9 @@ print_full_info (const char *username)
       {
 	if (i > 0)
 	  putchar (',');
-	printf ("%u", (unsigned) groups[i]);
+	printf ("%lu", (unsigned long int) groups[i]);
 	grp = getgrgid (groups[i]);
-	if (grp == NULL)
-	  problems++;
-	else
+	if (grp)
 	  printf ("(%s)", grp->gr_name);
       }
     free (groups);
