@@ -1,11 +1,12 @@
 # -*-perl-*-
 # Test "sort".
 
-# Copyright (C) 1996-1999, 2001-2007 Free Software Foundation, Inc.
+# Copyright (C) 1996, 1997, 1998, 1999, 2001, 2002, 2003, 2004, 2005,
+# 2006 Free Software Foundation, Inc.
 
-# This program is free software: you can redistribute it and/or modify
+# This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
-# the Free Software Foundation, either version 3 of the License, or
+# the Free Software Foundation; either version 2 of the License, or
 # (at your option) any later version.
 
 # This program is distributed in the hope that it will be useful,
@@ -14,7 +15,9 @@
 # GNU General Public License for more details.
 
 # You should have received a copy of the GNU General Public License
-# along with this program.  If not, see <http://www.gnu.org/licenses/>.
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+# 02110-1301, USA.
 
 package Test;
 require 5.002;
@@ -48,8 +51,6 @@ my @tv = (
 ["02a", '-c', "A\nB\nC\n", '', 0],
 ["02b", '-c', "A\nC\nB\n", '', 1],
 ["02c", '-c -k1,1', "a\na b\n", '', 0],
-["02d", '-C', "A\nB\nC\n", '', 0],
-["02e", '-C', "A\nC\nB\n", '', 1],
 # This should fail because there are duplicate keys
 ["02m", '-cu', "A\nA\n", '', 1],
 ["02n", '-cu', "A\nB\n", '', 0],
@@ -271,23 +272,9 @@ my @tv = (
 ["incompat2", '-fR', '', '', 2],
 ["incompat3", '-dfgiMnR', '', '', 2],
 ["incompat4", '-c -o /dev/null', '', '', 2],
-["incompat5", '-C -o /dev/null', '', '', 2],
-["incompat6", '-cC', '', '', 2],
 
 # -t '\0' is accepted, as of coreutils-5.0.91
 ['nul-tab', "-k2,2 -t '\\0'", "a\0z\01\nb\0y\02\n", "b\0y\02\na\0z\01\n", 0],
-
-["bigfield", '-k 340282366920938463463374607431768211456',
- "2\n1\n", "1\n2\n", 0],
-
-# Using an old-style key-specifying option like +1 with an invalid
-# ordering-option character would cause sort to try to free an invalid
-# (non-malloc'd) pointer.  This bug affects coreutils-6.5 through 6.9.
-['obs-inval', '+1x', '', '', 2],
-
-# Exercise the code that enlarges the line buffer.  See the thread here:
-# http://thread.gmane.org/gmane.comp.gnu.coreutils.bugs/11006
-['realloc-buf', '-S1', 'a'x4000 ."\n", 'a'x4000 ."\n", 0],
 );
 
 sub test_vector
